(* $Id: translate.ml,v 1.4 1996/12/03 14:12:45 garrigue Exp $ *)

#open"misc";;
#open"common";;
#open"syntax";;
#open"types";;
#open"define";;
#open"untyped";;

let rec pattern pat =
  match pat.sp_desc with
    SPid s ->
      begin try
	let info = hashtbl__find constructors s in
	if info.ci_size <> 0 then
	  failwith ("In translate__pattern, " ^ s ^
		    " : need arguments");
	UPblock (info.ci_tag, []), []
      with Not_found ->
	let id = new_id s in UPid id, [s,id]
      end
  | SPconst c -> UPconst c, []
  | SPtuple l ->
      let ul, bnds = split (map pattern l) in
      UPblock (0, ul), flatten bnds
  | SParray l ->
      let ul, bnds = split (map pattern l) in
      UPblock (0, ul), flatten bnds
  | SPconstr (s,pat) ->
      begin try
	let info = hashtbl__find constructors s in
	if info.ci_size = 0 then
	  failwith ("In translate__pattern, " ^ s ^
		    " : cannot be applied")
	else if info.ci_size = 1 then
	  let upat, bnds = pattern pat in
	  UPblock (info.ci_tag, [upat]), bnds
	else match pat.sp_desc with
	  SPtuple l when list_length l = info.ci_size ->
	    let upatl, bnds = split (map pattern l) in
	    UPblock (info.ci_tag, upatl), flatten bnds
	| _ ->
	    failwith
	    ("In translate__pattern, " ^ s ^ " : wrong number of arguments")
      with
	Not_found ->
	  failwith ("In translate__pattern, " ^ s ^ " : unkwnown constructor")
      end
  | SPrecord l ->
      let info = hashtbl__find labels (fst (hd l)) in
      let arr = make_vect info.li_total UPany in
      let bnds = map
        (fun (s, p1) ->
	  let up1, bnds = pattern p1
	  and info = hashtbl__find labels s in
	  arr.(info.li_index) <- up1;
	  bnds)
        l in
      (UPblock (0, list_of_vect arr), flatten bnds)
  | SPany -> UPany, []
  | SPtype (pat, _) -> pattern pat
;;

let current_def = ref "";;

let rec command bnds cmd =
  match cmd.sc_desc with
    SEexpr e ->
      [UEexpr (expression bnds e)], bnds
  | SEval l ->
      let patl, el = split l in
      let uel = map (expression bnds) el
      and upatl, new_bnds = split (map pattern patl) in
      map2 (fun pat (upat, ue) -> UEval (upat, ue, !current_def))
	   patl (combine (upatl,uel)),
      flatten new_bnds @ bnds
  | SEfun l ->
      let sl, el = split l in
      let ids = map new_id sl in
      let bnds = combine (sl, ids) @ bnds in
      let uel = map2
 	(fun s e ->
	  let old = !current_def in
	  if old = "" then current_def := s;
 	  let ue = expression bnds e in
	  current_def := old; ue)
 	sl el in
      [UEfun (combine (ids, uel))], bnds
  | SEvar l ->
      let sl, el = split l in
      let ids = map new_id sl
      and uel = map (expression bnds) el in
      let bnds = combine (sl, ids) @ bnds in
      [UEfun (combine (ids, uel))], bnds
  | STtype _ | SThide _ ->
      failwith "translate__command"

and expression bnds se =
  match se.se_desc with
    SEid s ->
      begin try
	let info = hashtbl__find constructors s in
	if info.ci_size <> 0 then
	  failwith ("In translate__expression, " ^ s ^
		    " : need arguments");
	UEblock (info.ci_tag, [])
      with Not_found -> try
	UEid (assoc s bnds)
      with Not_found ->
	failwith ("In translate__expression : unbound identifier " ^ s)
      end
  | SEconst c -> UEconst c
  | SEtuple l -> UEblock (0, map (expression bnds) l)
  | SEarray l -> UEblock (0, map (expression bnds) l)
  | SErecord l ->
      let arr = make_vect (list_length l) (UEblock(0,[])) in
      do_list
        (fun (s,e1) ->
	  let ue1 = expression bnds e1
	  and info = hashtbl__find labels s in
	  arr.(info.li_index) <- ue1)
        l;
      UEblock (0, list_of_vect arr)
  | SEapply ({se_desc = SEid s} as e0, [e]) ->
      begin try
	let info = hashtbl__find constructors s in
	if info.ci_size = 0 then
	  failwith ("In translate__expression, " ^ s ^
		    " : cannot be applied")
	else if info.ci_size = 1 then
	  UEblock (info.ci_tag, [expression bnds e])
	else begin match e.se_desc with
	  SEtuple l when list_length l = info.ci_size ->
	    let uel = map (expression bnds) l in
	    UEblock (info.ci_tag, uel)
	| _ ->
	    failwith ("In translate__expression, " ^ s ^
		      " : wrong number of arguments")
	end
      with Not_found ->
	UEapply (expression bnds e0, [expression bnds e])
      end
  | SEapply (e, l) ->
      UEapply (expression bnds e, map (expression bnds) l)
  | SEfunct (patl, e) ->
      let upatl, new_bnds = split (map pattern patl) in
      let ue = expression (flatten new_bnds @ bnds) e in
      let make_match pat = function
	  UPid id -> id, []
	| upat ->
	    let id = new_id "param" in
	    (id, [UEval (upat, UEid id, !current_def)])
      in
      let ids, vals = split (map2 make_match patl upatl) in
      UEfunct (ids, UEseq (flatten vals, ue))
  | SEseq l ->
      let cmds, last = split_last l in
      let ucmds, bnds =
	it_list
	  (fun (ucmds, bnds) cmd ->
	    let ucmd, bnds = command bnds cmd in
	    (ucmds @ ucmd, bnds))
	  ([], bnds) cmds in
      begin match last.sc_desc with
	SEexpr e -> UEseq (ucmds, expression bnds e)
      |	_ ->
	  let ucmd, _ = command bnds last in
	  UEseq (ucmds @ ucmd, UEblock (0,[]))
      end
  | SEcase (e, cases) ->
      let ue = expression bnds e in
      let ucases = map
	(fun (pat, e) ->
	  let upat, new_bnds = pattern pat in
	  (upat, expression (new_bnds @ bnds) e))
	cases in
      UEcase (ue, ucases, !current_def)
  | SEifthenelse (e1, e2, e3) ->
      UEifthenelse (expression bnds e1, expression bnds e2, expression bnds e3)
  | SEset (s, e) ->
      begin try
	UEset (assoc s bnds, expression bnds e)
      with Not_found ->
	failwith ("In translate__expression : unbound identifier " ^ s)
      end
  | SEgetfield (e1, s) ->
      let ue1 = expression bnds e1
      and info = hashtbl__find labels s in
      UEapply (UEid (assoc "$." bnds), [ue1;UEconst(Cint info.li_index)])
  | SEsetfield (e1, s, e2) ->
      let ue1 = expression bnds e1
      and ue2 = expression bnds e2
      and info = hashtbl__find labels s in
      UEapply (UEid (assoc "$.<-" bnds), [ue1;UEconst(Cint info.li_index);ue2])
  | SEfor (s, e1, dir, e2, e3) ->
      let ue1 = expression bnds e1
      and ue2 = expression bnds e2 in
      let id = new_id s in
      UEfor (id, ue1, dir, ue2, expression ((s,id)::bnds) e3)
  | SEwhile (e1, e2) ->
      UEwhile (expression bnds e1, expression bnds e2)
  | SEtype (e, _) ->
      expression bnds e
;;
